# CairoPlot.py
#
# Copyright (c) 2008 Rodrigo Moreira Araujo
#
# Author: Rodrigo Moreiro Araujo <alf.rodrigo@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import cairo
import math
import random
from exceptions import ValueError

def dot_line_plot(name, data, width, height, background = None, border = 0, axis = False, grid = False, h_legend = None, v_legend = None, h_bounds = None, v_bounds = None):
    '''
        Function to plot graphics using dots and lines.
        dot_line_plot (name, data, width, height, background = None, border = 0, axis = False, grid = False, h_legend = None, v_legend = None, h_bounds = None, v_bounds = None)

        Parameters

        name - Name of the desired output file, no need to input the .svg as it will be added at runtim;
        data - The list, list of lists or dictionary holding the data to be plotted;
        width, height - Dimensions of the output image;
        background - A 3 element tuple representing the rgb color expected for the background. If left None, a gray to white gradient will be generated;
        border - Distance in pixels of a square border into which the graphics will be drawn;
        axis - Whether or not the axis are to be drawn;
        grid - Whether or not the gris is to be drawn;
        h_legend, v_legend - lists of strings containing the horizontal and vertical legends for the axis;
        h_bounds, v_bounds - tuples containing the lower and upper value bounds for the data to be plotted.

        Examples of use

        teste_data = [0, 1, 3, 8, 9, 0, 10, 10, 2, 1]
        CairoPlot.dot_line_plot('teste', teste_data, 400, 300)
        
        teste_data_2 = {"john" : [10, 10, 10, 10, 30], "mary" : [0, 0, 3, 5, 15], "philip" : [13, 32, 11, 25, 2]}
        teste_h_legend = ["jan/2008", "feb/2008", "mar/2008", "apr/2008", "may/2008"]
        CairoPlot.dot_line_plot('teste2', teste_data_2, 400, 300, axis = True, grid = True, h_legend = teste_h_legend)
    '''

    surface = cairo.SVGSurface( name + '.svg', width, height)
    cr = cairo.Context(surface)
    
    cr.set_font_size(10)
    max_h_value = 0
    if h_legend:
        for item in h_legend :
            max_h_value = cr.text_extents(item)[2] if cr.text_extents(item)[2] > max_h_value else max_h_value
        v_border = max_h_value + border
    elif h_bounds:
        max_h_value = h_bounds[1]
        v_border = cr.text_extents(str(h_bounds[1]))[2] + border + 20
    else:
        h_bounds = (0,0)
        if type(data) == type({}):
            for key in data.keys():
                max_h_value = cr.text_extents(str(len(data[key])))[2] if cr.text_extents(str(len(data[key])))[2] > max_h_value else max_h_value
                h_bounds = (0,len(data[key])) if len(data[key]) > h_bounds[1] else h_bounds
        elif type(data[0]) == type([]):
            for item in data:
                max_h_value = cr.text_extents(str(len(item)))[2] if cr.text_extents(str(len(item)))[2] > max_h_value else max_h_value
                h_bounds = (0,len(item)) if len(item) > h_bounds[1] else h_bounds
        else:
            max_h_value = cr.text_extents(str(len(data)))[2]
            h_bounds = (0,len(data))
        v_border = max_h_value + border + 20

    max_v_value = 0
    if v_legend:
        for item in v_legend :
            max_v_value = cr.text_extents(item)[2] if cr.text_extents(item)[2] > max_v_value else max_v_value
        h_border = max_v_value + border + 10
        try:
            v_bounds = (0,int(v_legend[-1]))
        except ValueError:
            pass
            
    elif v_bounds:
        max_v_value = v_bounds[1]
        h_border = cr.text_extents(str(v_bounds[1]))[2] + border + 20
    else:
        v_bounds = (0,0)
        if type(data) == type({}):
            for key in data.keys():
                for value in data[key]:
                    max_v_value = cr.text_extents(str(value))[2] if cr.text_extents(str(value))[2] > max_v_value else max_v_value
                    v_bounds = (0, int(value)) if int(value) > v_bounds[1] else v_bounds
        elif type(data[0]) == type([]):
            for key in data:
                for value in key:
                    max_v_value = cr.text_extents(str(value))[2] if cr.text_extents(str(value))[2] > max_v_value else max_v_value
                    v_bounds = (0, int(value)) if int(value) > v_bounds[1] else v_bounds
        else:
            for value in data:
                max_v_value = cr.text_extents(str(value))[2] if cr.text_extents(str(value))[2] > max_v_value else max_v_value
                v_bounds = (0, int(value)) if int(value) > v_bounds[1] else v_bounds
        h_border = max_v_value + border + 20

    if background != None:
        cr.set_source_rgb(background[0], background[1], background[2])
        cr.rectangle(0,0,width,height)
        cr.fill()
        cr.set_source_rgb(0.5, 0.5, 0.5)
        cr.set_line_width(0.5)
        cr.rectangle(border,border,width-2*border,height-2*border)
        cr.stroke()
    else:
        linear = cairo.LinearGradient(width/2, 0, width/2, height)
        linear.add_color_stop_rgb(0,1.0,1.0,1.0)
        linear.add_color_stop_rgb(1.0,0.9,0.9,0.9)
        cr.set_source(linear)
        cr.rectangle(0,0, width, height)
        cr.fill()
        cr.set_source_rgb(0.5, 0.5, 0.5)
        cr.set_line_width(0.5)
        cr.rectangle(border,border,width-2*border,height-2*border)
        cr.stroke()
   
    if axis :
        cr.set_source_rgb(0.5, 0.5, 0.5)
        cr.move_to(h_border, height - v_border)
        cr.line_to(h_border, v_border)
        cr.stroke()
        cr.move_to(h_border, height - v_border)
        cr.line_to(width - h_border, height - v_border)
        cr.stroke()

    cr.set_font_size(8)
    if h_legend != None:
        h_step = (width - 2*h_border)/len(h_legend)
        x = h_border
        for item in h_legend:
            cr.set_source_rgb(0.0, 0.0, 0.0)
            w = cr.text_extents(item)[2]
            cr.move_to(x,height - v_border + 10)
            cr.rotate(math.pi/2.5)
            cr.show_text(item)
            cr.rotate(-math.pi/2.5)
            if grid and x != h_border:
                cr.set_source_rgb(0.8, 0.8, 0.8)
                cr.move_to(x, height - v_border)
                cr.line_to(x, v_border)
                cr.stroke()
            x += h_step
    else:
        h_step = (width - 2*h_border)/h_bounds[1]
        x = h_border
        value = 0
        while value < h_bounds[1]:
            cr.set_source_rgb(0.0, 0.0, 0.0)
            w = cr.text_extents(str(value))[2]
            cr.move_to(x,height - v_border + 10)
            cr.show_text(str(value))
            if grid and x != h_border:
                cr.set_source_rgb(0.8, 0.8, 0.8)
                cr.move_to(x, height - v_border)
                cr.line_to(x, v_border)
                cr.stroke()
            x += h_step
            value += 1

    if v_legend != None:
        v_step = (height - 2*v_border)/len(v_legend)
        y = height - v_border
        for item in v_legend:
            cr.set_source_rgb(0.0, 0.0, 0.0)
            w = cr.text_extents(item)[2]
            cr.move_to(h_border - w - 5,y)
            cr.show_text(item)
            if grid and y != height - v_border:
                cr.set_source_rgb(0.8, 0.8, 0.8)
                cr.move_to(h_border, y)
                cr.line_to(width - h_border, y)
                cr.stroke()
            y -= v_step
    else:
        v_step = (height - 2*v_border)/10.0
        y = height - v_border
        value = 0
        while value <= v_bounds[1]:
            cr.set_source_rgb(0.0, 0.0, 0.0)
            w = cr.text_extents(str(value))[2]
            cr.move_to(h_border - w - 5,y)
            cr.show_text(str(value))
            if grid and value != 0:
                cr.set_source_rgb(0.8, 0.8, 0.8)
                cr.move_to(h_border, y)
                cr.line_to(width - h_border, y)
                cr.stroke()
            y -= v_step
            value += v_bounds[1]/10.0

    if type(data) == type({}):
        max_value = 0
        max_key = ''
        if v_bounds :
            max_value = v_bounds[1]
        else:
            for key in data.keys():
                max_key = key if len(key) > len(max_key) else max_key
                for value in data[key]:
                    max_value = int(value) if int(value) > max_value else max_value
        print max_value
        for key in data.keys():
            horizontal_step = (width - 2*h_border)/len(data[key])
            vertical_step = (height - 2*v_border)/max_value
            cr.set_source_rgb(random.random(), random.random(), random.random())

            w,h = cr.text_extents(key)[2],cr.text_extents(key)[3]
            cr.move_to(width - h_border - horizontal_step - cr.text_extents(max_key)[2] + 10, 
                       height - v_border - int(data[key][-1])*vertical_step + h/2)
            cr.show_text(key)

            x = h_border
            last = None
            for value in data[key]:
                if last != None:
                    cr.move_to(x-horizontal_step, height - v_border - int(last)*vertical_step)
                    cr.line_to(x, height - v_border - int(value)*vertical_step)
                    cr.set_line_width(1)
                    cr.stroke()
                cr.arc(x, height - v_border - int(value)*vertical_step, 3, 0, 2*math.pi)
                cr.fill()
                x += horizontal_step
                last = value
    elif type(data) == type([]):
        if type(data[0]) == type([]):
            max_value = 0
            if v_bounds :
                max_value = v_bounds[1]
            else:
                for key in data:
                    for value in key:
                        max_value = int(value) if int(value) > max_value else max_value
            horizontal_step = (width - 2*h_border)/len(data[key])
            vertical_step = (height - 2*v_border)/max_value
            last = None
            for key in data:
                cr.set_source_rgb(random.random(), random.random(), random.random())
                x = h_border
                last = None
                for value in key:
                    if last != None:
                        cr.move_to(x-horizontal_step, height - v_border - int(last)*vertical_step)
                        cr.line_to(x, height - v_border - int(value)*vertical_step)
                        cr.set_line_width(1)
                        cr.stroke()
                    cr.arc(x, height - v_border - int(value)*vertical_step, 3, 0, 2*math.pi)
                    cr.fill()
                    x += horizontal_step
                    last = value
        else:
            max_value = 0
            if v_bounds :
                max_value = v_bounds[1]
            else:
                for key in data:
                    max_value = int(value) if int(value) > max_value else max_value
            horizontal_step = (width - 2*h_border)/len(data)
            vertical_step = (height - 2*v_border)/max_value
            last = None
            x = h_border
            cr.set_source_rgb(random.random(), random.random(), random.random())
            for value in data:
                if last != None:
                    cr.move_to(x-horizontal_step, height - v_border - int(last)*vertical_step)
                    cr.line_to(x, height - v_border - int(value)*vertical_step)
                    cr.set_line_width(1)
                    cr.stroke()
                cr.arc(x, height - v_border - int(value)*vertical_step, 3, 0, 2*math.pi)
                cr.fill()
                x += horizontal_step
                last = value



def pizza_plot(name, data, width, height, background = None):

    '''
        Function to plot pizza graphics.
        pizza_plot(name, data, width, height, background = None)

        Parameters
        
        name - Name of the desired output file, no need to input the .svg as it will be added at runtim;
        data - The list, list of lists or dictionary holding the data to be plotted;
        width, height - Dimensions of the output image;
        background - A 3 element tuple representing the rgb color expected for the background. If left None, a gray to white gradient will be generated;

        Examples of use
        
        teste_data = {"john" : 123, "mary" : 489, "philip" : 890 , "suzy" : 235}
        CairoPlot.pizza_plot("pizza_teste", teste_data, 500, 500)
        
    '''

    surface = cairo.SVGSurface(name + '.svg', width, height)
    print width, height
    cr = cairo.Context(surface)

    if background != None:
        cr.set_source_rgb(background[0], background[1], background[2])
        cr.rectangle(0,0,width,height)
        cr.fill()
    else:
        linear = cairo.LinearGradient(width/2, 0, width/2, height)
        linear.add_color_stop_rgb(0,1.0,1.0,1.0)
        linear.add_color_stop_rgb(1.0,0.9,0.9,0.9)
        cr.set_source(linear)
        cr.rectangle(0,0, width, height)
        cr.fill()
 
    angle = 0
    next_angle = 0
    cr.set_line_width(2.0)
    x0 = width/2
    y0 = height/2
    n = 0
    for key in data.keys():
        n += data[key]

    for key in data.keys():
        next_angle = angle + 2.0*math.pi*data[key]/n
    
        radius = width/3 if width < height else height/3
        print radius
        cr.set_source_rgb(random.random(), random.random(), random.random())
    
        w = cr.text_extents(key)[2]
        if (angle + next_angle)/2 < math.pi/2 or (angle + next_angle)/2 > 3*math.pi/2:
            cr.move_to(x0 + (radius+10)*math.cos((angle+next_angle)/2), y0 + (radius+10)*math.sin((angle+next_angle)/2) )
        else:
            cr.move_to(x0 + (radius+10)*math.cos((angle+next_angle)/2) - w, y0 + (radius+10)*math.sin((angle+next_angle)/2) )
        cr.show_text(key)

        cr.move_to(x0,y0)
        cr.line_to(x0+radius*math.cos(angle), y0+radius*math.sin(angle))
        cr.arc(x0, y0, radius, angle, angle + 2.0*math.pi*data[key]/n)
        cr.line_to(x0,y0)
        cr.close_path()
        cr.fill()
        cr.set_source_rgb(1.0, 1.0, 1.0)
        cr.move_to(x0,y0)
        cr.line_to(x0+radius*math.cos(angle), y0+radius*math.sin(angle))
        cr.arc(x0, y0, radius, angle, angle + 2.0*math.pi*data[key]/n)
        cr.line_to(x0,y0)
        cr.close_path()
        cr.stroke()
        angle = next_angle


def drawRectangle(cr, x0, y0, x1, y1, color):
    mid = (x0+x1)/2
    linear = cairo.LinearGradient(mid,y0,mid,y1)
    linear.add_color_stop_rgb(0,3.5*color[0]/5.0, 3.5*color[1]/5.0, 3.5*color[2]/5.0)
    linear.add_color_stop_rgb(1,color[0],color[1],color[2])
    cr.set_source(linear)

    cr.arc(x0+5, y0+5, 5, 0, 2*math.pi)
    cr.arc(x1-5, y0+5, 5, 0, 2*math.pi)
    cr.arc(x0+5, y1-5, 5, 0, 2*math.pi)
    cr.arc(x1-5, y1-5, 5, 0, 2*math.pi)
    cr.rectangle(x0+5,y0,x1-x0-10,y1-y0)
    cr.rectangle(x0,y0+5,x1-x0,y1-y0-10)
    cr.fill()

def drawShadow(cr, x0, y0, x1, y1):
    shadow = 0.4
    h_mid = (x0+x1)/2
    v_mid = (y0+y1)/2
    h_linear_1 = cairo.LinearGradient(h_mid,y0-4,h_mid,y0+4)
    h_linear_2 = cairo.LinearGradient(h_mid,y1-4,h_mid,y1+4)
    v_linear_1 = cairo.LinearGradient(x0-4,v_mid,x0+4,v_mid)
    v_linear_2 = cairo.LinearGradient(x1-4,v_mid,x1+4,v_mid)
    radial_00 = cairo.RadialGradient(x0+4, y0+4, 0, x0+4, y0+4, 8)
    radial_01 = cairo.RadialGradient(x1-4, y0+4, 0, x1-4, y0+4, 8)
    radial_10 = cairo.RadialGradient(x0+4, y1-4, 0, x0+4, y1-4, 8)
    radial_11 = cairo.RadialGradient(x1-4, y1-4, 0, x1-4, y1-4, 8)


    h_linear_1.add_color_stop_rgba( 0, 0, 0, 0, 0)
    h_linear_1.add_color_stop_rgba( 1, 0, 0, 0, shadow)
    h_linear_2.add_color_stop_rgba( 0, 0, 0, 0, shadow)
    h_linear_2.add_color_stop_rgba( 1, 0, 0, 0, 0)
    v_linear_1.add_color_stop_rgba( 0, 0, 0, 0, 0)
    v_linear_1.add_color_stop_rgba( 1, 0, 0, 0, shadow)
    v_linear_2.add_color_stop_rgba( 0, 0, 0, 0, shadow)
    v_linear_2.add_color_stop_rgba( 1, 0, 0, 0, 0)
    cr.set_source(h_linear_1)
    #cr.set_source_rgb(0,0,1)
    cr.rectangle(x0+4,y0-4,x1-x0-8,8)
    cr.fill()
    cr.set_source(h_linear_2)
    #cr.set_source_rgb(0,0,1)
    cr.rectangle(x0+4,y1-4,x1-x0-8,8)
    cr.fill()
    cr.set_source(v_linear_1)
    #cr.set_source_rgb(0,0,1)
    cr.rectangle(x0-4,y0+4,8,y1-y0-8)
    cr.fill()
    cr.set_source(v_linear_2)
    #cr.set_source_rgb(0,0,1)
    cr.rectangle(x1-4,y0+4,8,y1-y0-8)
    cr.fill()

    radial_00.add_color_stop_rgba(0, 0, 0, 0, shadow)
    radial_00.add_color_stop_rgba(1, 0, 0, 0, 0)
    radial_01.add_color_stop_rgba(0, 0, 0, 0, shadow)
    radial_01.add_color_stop_rgba(1, 0, 0, 0, 0)
    radial_10.add_color_stop_rgba(0, 0, 0, 0, shadow)
    radial_10.add_color_stop_rgba(1, 0, 0, 0, 0)
    radial_11.add_color_stop_rgba(0, 0, 0, 0, shadow)
    radial_11.add_color_stop_rgba(1, 0, 0, 0, 0)
    #cr.set_source_rgb(0,0,1)
    cr.set_source(radial_00)
    cr.move_to(x0+4,y0+4)
    cr.line_to(x0,y0+4)
    cr.arc(x0+4, y0+4, 8, math.pi, 3*math.pi/2)
    cr.line_to(x0+4,y0+4)
    cr.close_path()
    cr.fill()
    #cr.set_source_rgb(0,0,1)
    cr.set_source(radial_01)
    cr.move_to(x1-4,y0+4)
    cr.line_to(x1-4,y0)
    cr.arc(x1-4, y0+4, 8, 3*math.pi/2, 2*math.pi)
    cr.line_to(x1-4,y0+4)
    cr.close_path()
    cr.fill()
    #cr.set_source_rgb(0,0,0)
    cr.set_source(radial_10)
    cr.move_to(x0+4,y1-4)
    cr.line_to(x0+4,y1)
    cr.arc(x0+4, y1-4, 8, math.pi/2, math.pi)
    cr.line_to(x0+4,y1-4)
    cr.close_path()
    cr.fill()
    #cr.set_source_rgb(0,0,0)
    cr.set_source(radial_11)
    cr.move_to(x1-4,y1-4)
    cr.line_to(x1,y1-4)
    cr.arc(x1-4, y1-4, 8, 0, math.pi/2)
    cr.line_to(x1-4,y1-4)
    cr.close_path()
    cr.fill()

def gantt_chart(name, pieces, width, height, h_legend, v_legend, colors):

    '''
        Function to generate Gantt Diagrams.
        gantt_chart(name, pieces, width, height, h_legend, v_legend, colors):

        Parameters
        
        name - Name of the desired output file, no need to input the .svg as it will be added at runtim;
        pieces - A list defining the spaces to be drawn. The user must pass, for each line, the index of its start and the index of its end. If a line must have two or more spaces, they must be passed inside a list;
        width, height - Dimensions of the output image;
        h_legend - A list of names for each of the vertical lines;
        v_legend - A list of names for each of the horizontal spaces;
        colors - List containing the colors expected for each of the horizontal spaces

        Example of use

        pieces = [ (0.5,5.5) , [(0,4),(6,8)] , (5.5,7) , (7,8)]
        h_legend = [ 'teste01', 'teste02', 'teste03', 'teste04']
        v_legend = [ '0001', '0002', '0003', '0004', '0005', '0006', '0007', '0008', '0009', '0010' ]
        colors = [ (1.0, 0.0, 0.0), (1.0, 0.7, 0.0), (1.0, 1.0, 0.0), (0.0, 1.0, 0.0) ]
        CairoPlot.gantt_chart('gantt_teste', pieces, 600, 300, h_legend, v_legend, colors)
        
    '''

    surface = cairo.SVGSurface(name + '.svg', width, height)
    cr = cairo.Context(surface)
    cr.set_source_rgb(1.0, 1.0, 1.0)
    cr.rectangle(0,0,width,height)
    cr.fill()
    cr.set_font_size(0.02*width)
    max_word = ''
    for word in h_legend:
        max_word = word if word != None and len(word) > len(max_word) else max_word
    h_border = 100 + cr.text_extents(max_word)[2]
    horizontal_step = (width-h_border)/len(v_legend)
    vertical_step = height/(len(h_legend) + 1)
    v_border = vertical_step

    for line in pieces:
        linear = cairo.LinearGradient(width/2, v_border + pieces.index(line)*vertical_step, width/2, v_border + (pieces.index(line) + 1)*vertical_step)
        linear.add_color_stop_rgb(0,1.0,1.0,1.0)
        linear.add_color_stop_rgb(1.0,0.9,0.9,0.9)
        cr.set_source(linear)
        cr.rectangle(0,v_border + pieces.index(line)*vertical_step,width,vertical_step)
        cr.fill()

    cr.set_font_size(0.015*width)
    cr.set_source_rgb(0.7, 0.7, 0.7)
    cr.set_dash((1,0,0,0,0,0,1))
    cr.set_line_width(0.5)
    for word in v_legend:
        w,h = cr.text_extents(word)[2], cr.text_extents(word)[3]
        cr.move_to(h_border + v_legend.index(word)*horizontal_step-w/2, vertical_step/2)
        cr.show_text(word)
        cr.move_to(h_border + v_legend.index(word)*horizontal_step, vertical_step/2 + h)
        cr.line_to(h_border + v_legend.index(word)*horizontal_step, height)
    cr.stroke()

    cr.set_font_size(0.02*width)
    for line in pieces:
        word = h_legend[pieces.index(line)]
        if word != None:
            cr.set_source_rgb(0.5, 0.5, 0.5)
            w,h = cr.text_extents(word)[2], cr.text_extents(word)[3]
            cr.move_to(40,v_border + pieces.index(line)*vertical_step + vertical_step/2 + h/2)
            cr.show_text(word)

        if type(line) == type([]):
            for space in line:
                drawShadow(cr, h_border + space[0]*horizontal_step, v_border + pieces.index(line)*vertical_step + vertical_step/4.0, 
                               h_border + space[1]*horizontal_step, v_border + pieces.index(line)*vertical_step + 3.0*vertical_step/4.0)
                drawRectangle(cr, h_border + space[0]*horizontal_step, v_border + pieces.index(line)*vertical_step + vertical_step/4.0, 
                                  h_border + space[1]*horizontal_step, v_border + pieces.index(line)*vertical_step + 3.0*vertical_step/4.0, colors[pieces.index(line)])
        else:
            space = line
            drawShadow(cr, h_border + space[0]*horizontal_step, v_border + pieces.index(line)*vertical_step + vertical_step/4.0, 
                           h_border + space[1]*horizontal_step, v_border + pieces.index(line)*vertical_step + 3.0*vertical_step/4.0)
            drawRectangle(cr, h_border + space[0]*horizontal_step, v_border + pieces.index(line)*vertical_step + vertical_step/4.0, 
                              h_border + space[1]*horizontal_step, v_border + pieces.index(line)*vertical_step + 3.0*vertical_step/4.0, colors[pieces.index(line)])

